package edu.stanford.nlp.tagger.maxent;

import java.util.Properties;
import edu.stanford.nlp.util.StringUtils;
import java.io.PrintStream;
import java.io.*;

/**
 * Reads and stores configuration information for a POS tagger.
 *  @author William Morgan
 *  @author Anna Rafferty
 */
public class TaggerConfig {

  private String model, file, search, lang, arch, debugPrefix, tokenizerFactory, delimiter, encoding, distSimPath, chineseDictionaryPath, xmlOutput;
  private boolean debug, tokenize, learnClosedClassTags, useDistSim, useChineseDict, verbose;
  private int iterations, rareWordThresh, minFeatureThresh, curWordMinFeatureThresh, rareWordMinFeatureThresh, veryCommonWordThresh,
              closedClassTagThreshold;
  private double sigmaSquared;
  private String[] openClassTags, closedClassTags, xmlInput;
  public boolean sgml;

  public enum Mode {
    TRAIN, TEST, TAG, CONVERT
  }
  private Mode mode = Mode.TRAIN;

  /* defaults */
  private static String
  SEARCH = "cg",
  DELIMITER = "/",
  TOKENIZE = "true",
  DEBUG = "false",
  ITERATIONS = "100",
  RARE_WORD_THRESH = String.valueOf(GlobalHolder.rareWordThresh),
  MIN_FEATURE_THRESH = String.valueOf(GlobalHolder.minFeatureThresh),
  CUR_WORD_MIN_FEATURE_THRESH = String.valueOf(GlobalHolder.curWordMinFeatureThresh),
  RARE_WORD_MIN_FEATURE_THRESH = String.valueOf(GlobalHolder.rareWordMinFeatureThresh),
  VERY_COMMON_WORD_THRESH = String.valueOf(GlobalHolder.veryCommonWordThresh),
  SIGMA_SQUARED = String.valueOf(0.5),
  ENCODING = "UTF-8",
  LEARN_CLOSED_CLASS = "false",
  CLOSED_CLASS_THRESHOLD = String.valueOf(TTags.CLOSED_TAG_THRESHOLD),
  USE_DIST_SIM = "false",
  DIST_SIM_PATH = "/u/nlp/data/pos_tags_are_useless/egw.bnc.200.pruned",
  CHINESE_DICT_PATH = "/u/nlp/data/pos-tagger/dictionary",
  VERBOSE = "false",
  SGML = "false";


  public TaggerConfig(String[] args) throws IOException {
    Properties props = StringUtils.argsToProperties(args);

    if (props.getProperty("") != null) {
      throw new RuntimeException("unknown argument(s): \"" + props.getProperty("") + "\"");
    }

    if (props.getProperty("genprops") != null) {
      printGenProps();
      System.exit(0);
    }

    //Get the path to the model (or the path where you'd like to save the model); this is necessary for training, testing, and tagging
    model = props.getProperty("model", "").trim();
    if (model.equals("")) {
      throw new RuntimeException("'model' parameter must be specified");
    }

    /* Try and use the default properties from the model */
    Properties modelProps = new Properties();
    String mpfn = model + ".props";
    try {
      System.err.println("Trying to load default properties from model config " + mpfn);
      modelProps.load(new BufferedInputStream(new FileInputStream(mpfn)));
    } catch (FileNotFoundException e) {
      System.err.println("No such model config file found.");
    }

    //Figure out what mode we're in
    if(props.containsKey("trainFile")) {
      //Training mode
      mode = Mode.TRAIN;
      file = props.getProperty("trainFile", "").trim();
    } else if(props.containsKey("testFile")) {
      //Testing mode
      mode = Mode.TEST;
      file = props.getProperty("testFile", "").trim();
    } else if(props.containsKey("textFile")) {
      //Tagging mode
      mode = Mode.TAG;
      file = props.getProperty("textFile", "").trim();
    } else if(props.containsKey("convertToSingleFile")) {
      mode = Mode.CONVERT;
      file = props.getProperty("convertToSingleFile").trim();
    } else {
      mode = Mode.TAG;
      file = "stdin";
    }

    search = props.getProperty("search", modelProps.getProperty("search", SEARCH)).trim().toLowerCase();
    if ( ! search.equals("cg") && ! search.equals("iis")) {
      throw new RuntimeException("'search' must be one of 'cg' or 'iis': " + search);
    }

    sigmaSquared = Double.parseDouble(props.getProperty("sigmaSquared", modelProps.getProperty("sigmaSquared", SIGMA_SQUARED)));

    delimiter = props.getProperty("delimiter", modelProps.getProperty("delimiter", DELIMITER));
    TestSentence.delimiter = ((delimiter == null || delimiter.equals("")) ? TestSentence.delimiter : delimiter);

    iterations = Integer.parseInt(props.getProperty("iterations", modelProps.getProperty("iterations", ITERATIONS)));
    rareWordThresh = Integer.parseInt(props.getProperty("rareWordThresh", modelProps.getProperty("rareWordThresh", RARE_WORD_THRESH)));
    minFeatureThresh = Integer.parseInt(props.getProperty("minFeatureThresh", modelProps.getProperty("minFeatureThresh", MIN_FEATURE_THRESH)));
    curWordMinFeatureThresh = Integer.parseInt(props.getProperty("curWordMinFeatureThresh", modelProps.getProperty("curWordMinFeatureThresh", CUR_WORD_MIN_FEATURE_THRESH)));
    rareWordMinFeatureThresh = Integer.parseInt(props.getProperty("rareWordMinFeatureThresh", modelProps.getProperty("rareWordMinFeatureThresh", RARE_WORD_MIN_FEATURE_THRESH)));
    veryCommonWordThresh = Integer.parseInt(props.getProperty("veryCommonWordThresh", modelProps.getProperty("veryCommonWordThresh", VERY_COMMON_WORD_THRESH)));



    lang = props.getProperty("lang", modelProps.getProperty("lang", ""));
    String ocs = props.getProperty("openClassTags", modelProps.getProperty("openClassTags", null));

    if((ocs == null) || ocs.equals("") || ocs.equals("true")) /* necessary because of peculiarities of java */
      openClassTags = new String[0];
    else
      openClassTags = ocs.split("\\s+");

    String ccs = props.getProperty("closedClassTags", modelProps.getProperty("closedClassTags", null));

    if((ccs == null) || ccs.equals("") || ccs.equals("true")) /* necessary because of peculiarities of java */
      closedClassTags = new String[0];
    else
      closedClassTags = ccs.split("\\s+");
    learnClosedClassTags = Boolean.parseBoolean(props.getProperty("learnClosedClassTags", modelProps.getProperty("learnClosedClassTags", LEARN_CLOSED_CLASS)));

    closedClassTagThreshold = Integer.parseInt(props.getProperty("closedClassTagThreshold", modelProps.getProperty("closedClassTagThreshold", CLOSED_CLASS_THRESHOLD)));

    arch = props.getProperty("arch", modelProps.getProperty("arch", ""));
    tokenize = Boolean.parseBoolean(props.getProperty("tokenize", modelProps.getProperty("tokenize", TOKENIZE)));
    tokenizerFactory = props.getProperty("tokenizerFactory", modelProps.getProperty("tokenizerFactory", ""));

    debugPrefix = props.getProperty("debugPrefix", "");
    debug = Boolean.parseBoolean(props.getProperty("debug", DEBUG));

    encoding = props.getProperty("encoding", modelProps.getProperty("encoding", ENCODING));
    useDistSim = Boolean.parseBoolean(props.getProperty("useDistSim", modelProps.getProperty("useDistSim", USE_DIST_SIM)));
    distSimPath = props.getProperty("distSimPath", modelProps.getProperty("distSimPath", DIST_SIM_PATH));
    sgml = Boolean.parseBoolean(props.getProperty("sgml", modelProps.getProperty("sgml", SGML)));
    verbose = Boolean.parseBoolean(props.getProperty("verbose", modelProps.getProperty("verbose", VERBOSE)));


    String xmlInputProp = props.getProperty("xmlInput", modelProps.getProperty("xmlInput", null));
    if((xmlInputProp == null) || xmlInputProp.equals("")) /* necessary because of peculiarities of java */
      xmlInput = new String[0];
    else
      xmlInput = xmlInputProp.split("\\s+");
    xmlOutput = props.getProperty("xmlOutput", modelProps.getProperty("xmlOutput", ""));


    //Chinese properties
    chineseDictionaryPath = props.getProperty("chineseDictionaryPath",modelProps.getProperty("chineseDictionaryPath", CHINESE_DICT_PATH)).trim();
    useChineseDict = Boolean.parseBoolean(props.getProperty("useChineseDictionaries", modelProps.getProperty("useChineseDictionaries", "false")));

  }


  public String getModel() { return model; }

  public String getFile() { return file; }

  public String getSearch() { return search; }

  public double getSigmaSquared() { return sigmaSquared; }

  public int getIterations() { return iterations; }

  public int getRareWordThresh() { return rareWordThresh; }

  public int getMinFeatureThresh() { return minFeatureThresh; }

  public int getCurWordMinFeatureThresh() { return curWordMinFeatureThresh; }

  public int getRareWordMinFeatureThresh() { return rareWordMinFeatureThresh; }

  public int getVeryCommonWordThresh() { return veryCommonWordThresh; }

  public String getLang() { return lang; }

  public String[] getOpenClassTags() { return openClassTags; }

  public String[] getClosedClassTags() { return closedClassTags; }

  public boolean getLearnClosedClassTags() { return learnClosedClassTags; }

  public int getClosedTagThreshold() { return closedClassTagThreshold; }

  public String getArch() { return arch; }

  public boolean getDebug() { return debug; }

  public String getDebugPrefix() { return debugPrefix; }

  public String getTokenizerFactory() { return tokenizerFactory; }

  public String getDelimiter() { return delimiter; }

  public boolean getTokenize() { return tokenize; }

  public String getEncoding() { return encoding; }

  public boolean getUseDistSim() { return useDistSim; }

  public String getDistSimPath() { return distSimPath; }

  public String getChineseDictionaryPath() { return chineseDictionaryPath; }

  public boolean getUseChineseDict() { return useChineseDict; }

  public String[] getXMLInput() { return xmlInput; }

  public String getXMLOutput() { return xmlOutput; }


  public boolean getVerbose() { return verbose; }


  public void dump() { dump(System.err); }

  public void dump(PrintStream stream) {
    stream.println("                   model = " + model);
    if(mode == Mode.TRAIN)
      stream.println("                    trainFile = " + file);
    if(mode == Mode.TAG)
      stream.println("                    textFile = " + file);
    if(mode == Mode.TEST)
      stream.println("                    testFile = " + file);
    stream.println("               delimiter = " + delimiter);
    stream.println("        tokenizerFactory = " + tokenizerFactory);
    stream.println("                  search = " + search);
    stream.println("            sigmaSquared = " + sigmaSquared);
    stream.println("              iterations = " + iterations);
    stream.println("                    lang = " + lang);
    stream.println("           openClassTags = " + StringUtils.join(openClassTags, " "));
    stream.println("         closedClassTags = " + StringUtils.join(closedClassTags, " "));
    stream.println("    learnClosedClassTags = " + learnClosedClassTags);
    stream.println(" closedClassTagThreshold = " + closedClassTagThreshold);

    stream.println("                    arch = " + arch);
    stream.println("                   debug = " + debug);
    stream.println("             debugPrefix = " + debugPrefix);
    stream.println("                encoding = " + encoding);
    stream.println("                tokenize = " + tokenize);
    stream.println("          rareWordThresh = " + rareWordThresh);
    stream.println("        minFeatureThresh = " + minFeatureThresh);
    stream.println(" curWordMinFeatureThresh = " + curWordMinFeatureThresh);
    stream.println("rareWordMinFeatureThresh = " + rareWordMinFeatureThresh);
    stream.println("    veryCommonWordThresh = " + veryCommonWordThresh);
    stream.println("              useDistSim = " + useDistSim);
    stream.println("             distSimPath = " + distSimPath);



  }

  /**
   * Prints out the automatically generated props file - in its own
   * method to make code above easier to read
   */
  private static void printGenProps() {
    System.out.println("## Sample properties file for maxent tagger. This file is used for three main");
    System.out.println("## operations: training, testing, and tagging. It may also be used to convert a");
    System.out.println("## To train or test a model, or to tag something, run:");
    System.out.println("##   java edu.stanford.nlp.tagger.maxent.MaxentTagger -prop <properties file>");

    System.out.println("## Arguments can be overridden on the commandline, e.g.:");
    System.out.println("##   java ....MaxentTagger -prop <properties file> -file /other/file ");
    System.out.println("");
    System.out.println("# Model file name (created at train time; used at tag and test time)");
    System.out.println("# (you can leave this blank and specify it on the commandline with -model)");
    System.out.println("# model = ");
    System.out.println("");

    System.out.println("# Path to file to be operated on (trained from, tested against, or tagged)");
    System.out.println("# Specify -textFile <filename> to tag text in the given file, -trainFile <filename> to");
    System.out.println("# to train a model using data in the given file, or -testFile <filname> to test your");
    System.out.println("# model using data in the given file.  Alternatively, you may specify");
    System.out.println("# -convertToSingleFile <filename> to save an old, multi-file model (specified at -model) to");
    System.out.println("# the new single file format.  The new model will be saved in the file filename. If you choose to");
    System.out.println("# convert an old file, you must specify the correct 'arch' parameter used to create the original model.");
    System.out.println("# trainFile = ");
    System.out.println("");

    System.out.println("# Delimiter character that separates word and pos tags (for both training");
    System.out.println("# and test data).");
    System.out.println("# delimiter = " + DELIMITER);
    System.out.println("");

    System.out.println("# Encoding format in which files are stored.  If left blank, UTF-8 is assumed.");
    System.out.println("# encoding = " + ENCODING);
    System.out.println("");


    System.out.println("######### parameters for tag and test operations #########");
    System.out.println("");
    System.out.println("# Class to use for tokenization. Default blank value means Penn Treebank");
    System.out.println("# tokenization.  If you'd like to just assume that tokenization has been done,");
    System.out.println("# and the input is whitespace-tokenized, use");
    System.out.println("# edu.stanford.nlp.process.WhitespaceTokenizer or set tokenize to false.");
    System.out.println("# tokenizerFactory = ");
    System.out.println("");

    System.out.println("# Whether to tokenize text for tag and test operations. Default is true.");
    System.out.println("# If false, your text must already be whitespace tokenized.");
    System.out.println("# tokenize = " + TOKENIZE);
    System.out.println("");

    System.out.println("# Write debugging information (words, top words, unknown words). Useful for");
    System.out.println("# error analysis. Default is false.");
    System.out.println("# debug = "+ DEBUG);
    System.out.println("");

    System.out.println("# Prefix for debugging output (if debug == true). Default is to use the");
    System.out.println("# filename from 'file'");
    System.out.println("# debugPrefix = ");
    System.out.println("");

    System.out.println("######### parameters for training  #########");
    System.out.println("");
    System.out.println("# model architecture:");
    System.out.println("# 'left3words', 'left3wordswordshapes', left5wordswordshapes', 'bidirectional', ");
    System.out.println("# 'bidirectionalwordshapes', 'bidirectionalwordshapes5words', 'generic', '[1357]words?', or 'sighan2005'.");
    System.out.println("# This determines the features extracted.  'generic' is language independent.");
    System.out.println("# arch = ");
    System.out.println("");

    System.out.println("# language (for list of open-class tags). currently we have 'english', 'arabic', 'german', 'chinese'");
    System.out.println("# or 'polish' predefined. alternatively, you can specify openClassTags or closedClassTags");
    System.out.println("# below (only ONE of these three options may be specified). 'english' means penn treebank tags. 'polish' means");
    System.out.println("# some tags that some guy on the internet once used. See TTags class for more information; 'english' is the");
    System.out.println("# most accurate of the predefined languages ('arabic' and 'german' in particular are not very well defined).");
    System.out.println("# lang = ");
    System.out.println("");

    System.out.println("# a space-delimited list of open-class parts of speech");
    System.out.println("# alternatively, you can specify language above to use a pre-defined list or specify the closed class tags (below)");
    System.out.println("# openClassTags = ");
    System.out.println("");

    System.out.println("# a space-delimited list of closed-class parts of speech");
    System.out.println("# alternatively, you can specify language above to use a pre-defined list or specify the open class tags (above)");
    System.out.println("# closedClassTags = ");
    System.out.println("");

    System.out.println("# A boolean indicating whether you would like the trained model to set POS tags as closed");
    System.out.println("# based on their frequency in training; default is false.  The frequency threshold can be set below. ");
    System.out.println("# This option is ignored if any of {openClassTags, closedClassTags, lang} are specified.");
    System.out.println("# learnClosedClassTags = ");
    System.out.println("");

    System.out.println("# Used only if learnClosedClassTags=true.  Tags that have fewer tokens than this threshold are");
    System.out.println("# considered closed in the trained model.");
    System.out.println("# closedClassTagThreshold = ");
    System.out.println("");

    System.out.println("# search method. 'cg' (conjugate gradient (default)) or 'iis' (improved ");
    System.out.println("# iterative scaling)");
    System.out.println("# search = " + SEARCH);
    System.out.println("");

    System.out.println("# for conjugate gradient search, sigma-squared smoothing/regularization");
    System.out.println("# parameter. if left blank, the default is 0.5, which is usually okay");
    System.out.println("# sigmaSquared = " + SIGMA_SQUARED);
    System.out.println("");

    System.out.println("# for improved iterative scaling, the number of iterations");
    System.out.println("# iterations = " + ITERATIONS);
    System.out.println("");

    System.out.println("# rare word threshold. words that occur less than this number of");
    System.out.println("# times are considered rare words.");
    System.out.println("# rareWordThresh = " + RARE_WORD_THRESH);
    System.out.println("");

    System.out.println("# minimum feature threshold. features whose history appears less");
    System.out.println("# than this number of times are ignored.");
    System.out.println("# minFeatureThresh = " + MIN_FEATURE_THRESH);
    System.out.println("");

    System.out.println("# current word feature threshold. words that occur more than this");
    System.out.println("# number of times will generate features with all of their occurring");
    System.out.println("# tags.");
    System.out.println("# curWordMinFeatureThresh = " + CUR_WORD_MIN_FEATURE_THRESH);
    System.out.println("");

    System.out.println("# rare word minimum feature threshold. features of rare words whose histories");
    System.out.println("# appear less than this times will be ignored.");
    System.out.println("# rareWordMinFeatureThresh = " + RARE_WORD_MIN_FEATURE_THRESH);
    System.out.println("");

    System.out.println("# very common word threshold. words that occur more than this number of");
    System.out.println("# times will form an equivalence class by themselves. ignored unless");
    System.out.println("# you are using equivalence classes.");
    System.out.println("# veryCommonWordThresh = " + VERY_COMMON_WORD_THRESH);

    System.out.println("# useDistSim: Distributional similarity classes can be an added source of information");
    System.out.println("# about your words. An English distsim file is included, or you can use your own.  Distsim");
    System.out.println("# feature will be added in addition to your other features.");
    System.out.println("# useDistSim = " + USE_DIST_SIM);

    System.out.println("# distSimPath: Path to the distributional similarity file.  The included English distsim file");
    System.out.println("# is egw.bnc.200.pruned, located in the main part of speech tagger directory.");
    System.out.println("# distSimPath = ");

  }

  public Mode getMode() {
    return mode;
  }

}
